/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2025 IndiScale GmbH <info@indiscale.com>
 * Copyright (C) 2025 Joscha Schmiedt <joscha@schmiedt.dev>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

package org.caosdb.server.grpc;

import com.google.protobuf.Timestamp;
import io.grpc.Status;
import io.grpc.StatusException;
import io.grpc.stub.StreamObserver;
import java.io.IOException;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicLong;
import org.apache.shiro.SecurityUtils;
import org.apache.shiro.authz.AuthorizationException;
import org.apache.shiro.authz.UnauthorizedException;
import org.apache.shiro.subject.Subject;
import org.caosdb.api.scripting.v1alpha1.ExecuteServerSideScriptRequest;
import org.caosdb.api.scripting.v1alpha1.ExecuteServerSideScriptResponse;
import org.caosdb.api.scripting.v1alpha1.NamedArgument;
import org.caosdb.api.scripting.v1alpha1.ServerSideScriptExecutionId;
import org.caosdb.api.scripting.v1alpha1.ServerSideScriptExecutionResult;
import org.caosdb.api.scripting.v1alpha1.ServerSideScriptingServiceGrpc.ServerSideScriptingServiceImplBase;
import org.caosdb.server.accessControl.AuthenticationUtils;
import org.caosdb.server.accessControl.OneTimeAuthenticationToken;
import org.caosdb.server.accessControl.SessionToken;
import org.caosdb.server.entity.FileProperties;
import org.caosdb.server.entity.Message;
import org.caosdb.server.scripting.ScriptingPermissions;
import org.caosdb.server.scripting.ServerSideScriptingCaller;
import org.caosdb.server.utils.ServerMessages;

/**
 * Main entry point for the server-side scripting service of the server's GRPC API.
 *
 * @author Joscha Schmiedt <joscha@schmiedt.dev>
 */
public class ServerSideScriptingServiceImpl extends ServerSideScriptingServiceImplBase {

  private ServerSideScriptingCaller caller;
  private static AtomicLong executionId = new AtomicLong(0);

  @Override
  public void executeServerSideScript(
      ExecuteServerSideScriptRequest request,
      io.grpc.stub.StreamObserver<org.caosdb.api.scripting.v1alpha1.ExecuteServerSideScriptResponse>
          responseObserver) {

    try {
      AuthInterceptor.bindSubject();
      ExecuteServerSideScriptResponse response = callScript(request);
      responseObserver.onNext(response);
      responseObserver.onCompleted();

    } catch (final Exception e) {
      ServerSideScriptingServiceImpl.handleException(responseObserver, e);
    }
  }

  private long incrementAndGetExecutionId() {
    return executionId.incrementAndGet();
  }

  /**
   * Converts the ExecuteServerSideScriptRequest to a command line argument list.
   *
   * @param request the request containing the script filename and arguments
   * @return the command line argument list
   * @throws IllegalArgumentException if the script filename is missing
   */
  public static ArrayList<String> request2CommandLine(ExecuteServerSideScriptRequest request)
      throws IllegalArgumentException {

    if (request.getScriptFilename().length() == 0) {
      throw new IllegalArgumentException("Script filename is missing in the request.");
    }

    ArrayList<String> commandLine =
        new ArrayList<>(
            1 + request.getPositionalArgumentsCount() + request.getNamedArgumentsCount());

    // first item is the script filename
    commandLine.add(request.getScriptFilename());

    // add named arguments
    for (NamedArgument namedArgument : request.getNamedArgumentsList()) {
      String name = namedArgument.getName();
      String value = namedArgument.getValue();
      commandLine.add(String.format("--%s=%s", name, value));
    }

    // add positional arguments
    for (String positionalArgument : request.getPositionalArgumentsList()) {
      commandLine.add(positionalArgument);
    }

    return commandLine;
  }

  /////////////////////////////////////////////////////////////////////////////////////////
  // TODO: isAnonymous, getUser and generateAuthToken are copy-pasted from
  // elsewhere. Consider refactoring.
  /////////////////////////////////////////////////////////////////////////////////////////
  private static boolean isAnonymous() {
    return AuthenticationUtils.isAnonymous(getUser());
  }

  private static Subject getUser() {
    Subject ret = SecurityUtils.getSubject();
    return ret;
  }

  private static String[] constructCallStringArray(ArrayList<String> commandLine) {
    return commandLine.toArray(new String[commandLine.size()]);
  }

  private static Object generateAuthToken(String call) {
    String purpose = "SCRIPTING:EXECUTE:" + call;
    Object authtoken = OneTimeAuthenticationToken.generateForPurpose(purpose, getUser());
    if (authtoken != null || isAnonymous()) {
      return authtoken;
    }
    return SessionToken.generate(getUser());
  }

  private static void checkExecutionPermission(Subject user, String call)
      throws AuthorizationException {
    user.checkPermission(ScriptingPermissions.PERMISSION_EXECUTION(call));
  }

  /////////////////////////////////////////////////////////////////////////////////////////

  public ExecuteServerSideScriptResponse callScript(ExecuteServerSideScriptRequest request)
      throws Message {

    long startTimeMillis = System.currentTimeMillis();
    Timestamp startTime = Timestamp.newBuilder().setSeconds(startTimeMillis / 1000).build();

    checkExecutionPermission(getUser(), request.getScriptFilename());

    String[] call = constructCallStringArray(request2CommandLine(request));

    Integer timeoutMs = request.getTimeoutMs() != 0 ? (int) request.getTimeoutMs() : -1;

    if (request.getScriptFilesCount() > 0) {
      throw new UnsupportedOperationException(
          "Uploading files for scripting via gRPC is not supported yet.");
    }

    // auth token
    Object authToken = request.getAuthToken();
    if (authToken == null) {
      authToken = generateAuthToken(request.getScriptFilename());
    }

    // invoke script
    caller =
        new ServerSideScriptingCaller(call, timeoutMs, new ArrayList<FileProperties>(), authToken);
    int returnCode = caller.invoke();

    // construct response
    String scriptExecutionId = String.valueOf(incrementAndGetExecutionId());
    String scriptFilename = request.getScriptFilename();
    ServerSideScriptExecutionResult result =
        ServerSideScriptExecutionResult.SERVER_SIDE_SCRIPT_EXECUTION_RESULT_SUCCESS;

    String stdout = "";
    try {
      stdout = caller.getStdOut();
    } catch (final IOException e) {
      stdout = "Error retreiving stdout";
      result = ServerSideScriptExecutionResult.SERVER_SIDE_SCRIPT_EXECUTION_RESULT_GENERAL_FAILURE;
    }

    String stderr = "";
    try {
      stderr = caller.getStdErr();
    } catch (final IOException e) {
      stderr = "Error retrieving stderr";
      result = ServerSideScriptExecutionResult.SERVER_SIDE_SCRIPT_EXECUTION_RESULT_GENERAL_FAILURE;
    }

    long endTimeMillis = System.currentTimeMillis();
    Timestamp endTime = Timestamp.newBuilder().setSeconds(endTimeMillis / 1000).build();
    int duration_ms = (int) (endTimeMillis - startTimeMillis);

    final ExecuteServerSideScriptResponse response =
        ExecuteServerSideScriptResponse.newBuilder()
            .setScriptExecutionId(
                ServerSideScriptExecutionId.newBuilder()
                    .setScriptExecutionId(scriptExecutionId)
                    .build())
            .setCall(scriptFilename)
            .setResult(result)
            .setReturnCode(returnCode)
            .setStdout(stdout)
            .setStderr(stderr)
            .setStartTime(startTime)
            .setEndTime(endTime)
            .setDurationMs(duration_ms)
            .build();
    return response;
  }

  public static void handleException(StreamObserver<?> responseObserver, Exception e) {
    String description = e.getMessage();
    if (description == null || description.isBlank()) {
      description = "Unknown Error. Please Report!";
    }
    if (e instanceof UnauthorizedException) {
      Subject subject = SecurityUtils.getSubject();
      if (AuthenticationUtils.isAnonymous(subject)) {
        responseObserver.onError(new StatusException(AuthInterceptor.PLEASE_LOG_IN));
        return;
      } else {
        responseObserver.onError(
            new StatusException(
                Status.PERMISSION_DENIED.withCause(e).withDescription(description)));
        return;
      }
    } else if (e == ServerMessages.ROLE_DOES_NOT_EXIST
        || e == ServerMessages.ACCOUNT_DOES_NOT_EXIST) {
      responseObserver.onError(
          new StatusException(Status.NOT_FOUND.withDescription(description).withCause(e)));
      return;
    }
    // TODO: SERVER_SIDE_DOES_NOT_EXIST, SERVER_SIDE_SCRIPT_NOT_EXECUTABLE,
    // SERVER_SIDE_SCRIPT_ERROR, SERVER_SIDE_SCRIPT_SETUP_ERROR,
    // SERVER_SIDE_SCRIPT_TIMEOUT, SERVER_SIDE_SCRIPT_MISSING_CALL
    e.printStackTrace();
    responseObserver.onError(
        new StatusException(Status.UNKNOWN.withDescription(description).withCause(e)));
  }
}
